@@ -115,15 +115,15 @@ hsh_prealloc(struct worker *wrk)
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 
-	if (wrk->nobjcore == NULL)
-		wrk->nobjcore = ObjNew(wrk);
-	CHECK_OBJ_NOTNULL(wrk->nobjcore, OBJCORE_MAGIC);
+	if (wrk->wpriv->nobjcore == NULL)
+		wrk->wpriv->nobjcore = ObjNew(wrk);
+	CHECK_OBJ_NOTNULL(wrk->wpriv->nobjcore, OBJCORE_MAGIC);
 
-	if (wrk->nobjhead == NULL) {
-		wrk->nobjhead = hsh_newobjhead();
+	if (wrk->wpriv->nobjhead == NULL) {
+		wrk->wpriv->nobjhead = hsh_newobjhead();
 		wrk->stats->n_objecthead++;
 	}
-	CHECK_OBJ_NOTNULL(wrk->nobjhead, OBJHEAD_MAGIC);
+	CHECK_OBJ_NOTNULL(wrk->wpriv->nobjhead, OBJHEAD_MAGIC);
 
 	if (hash->prep != NULL)
 		hash->prep(wrk);
@@ -151,29 +151,34 @@ HSH_Private(const struct worker *wrk)
 }
 
 /*---------------------------------------------------------------------*/
+
 void
-HSH_Cleanup(struct worker *wrk)
+HSH_Cleanup(const struct worker *wrk)
 {
 
-	if (wrk->nobjcore != NULL)
-		ObjDestroy(wrk, &wrk->nobjcore);
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(wrk->wpriv, WORKER_PRIV_MAGIC);
+	if (wrk->wpriv->nobjcore != NULL)
+		ObjDestroy(wrk, &wrk->wpriv->nobjcore);
 
-	if (wrk->nobjhead != NULL) {
-		Lck_Delete(&wrk->nobjhead->mtx);
-		FREE_OBJ(wrk->nobjhead);
-		wrk->nobjhead = NULL;
+	if (wrk->wpriv->nobjhead != NULL) {
+		CHECK_OBJ(wrk->wpriv->nobjhead, OBJHEAD_MAGIC);
+		Lck_Delete(&wrk->wpriv->nobjhead->mtx);
+		FREE_OBJ(wrk->wpriv->nobjhead);
 		wrk->stats->n_objecthead--;
 	}
-	if (wrk->nhashpriv != NULL) {
+	if (wrk->wpriv->nhashpriv != NULL) {
 		/* XXX: If needed, add slinger method for this */
-		free(wrk->nhashpriv);
-		wrk->nhashpriv = NULL;
+		free(wrk->wpriv->nhashpriv);
+		wrk->wpriv->nhashpriv = NULL;
 	}
 }
 
 void
 HSH_DeleteObjHead(const struct worker *wrk, struct objhead *oh)
 {
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
 
 	AZ(oh->refcnt);
 	assert(VTAILQ_EMPTY(&oh->objcs));
@@ -191,7 +196,7 @@ HSH_AddString(struct req *req, void *ctx, const char *str)
 	AN(ctx);
 	if (str != NULL) {
 		VSHA256_Update(ctx, str, strlen(str));
-		VSLb(req->vsl, SLT_Hash, "%s", str);
+		VSLbs(req->vsl, SLT_Hash, TOSTRAND(str));
 	} else
 		VSHA256_Update(ctx, &str, 1);
 }
@@ -285,6 +290,7 @@ HSH_Insert(struct worker *wrk, const void *digest, struct objcore *oc,
 	struct rush rush;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(wrk->wpriv, WORKER_PRIV_MAGIC);
 	AN(digest);
 	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 	AN(ban);
@@ -295,8 +301,8 @@ HSH_Insert(struct worker *wrk, const void *digest, struct objcore *oc,
 
 	hsh_prealloc(wrk);
 
-	AN(wrk->nobjhead);
-	oh = hash->lookup(wrk, digest, &wrk->nobjhead);
+	AN(wrk->wpriv->nobjhead);
+	oh = hash->lookup(wrk, digest, &wrk->wpriv->nobjhead);
 	CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
 	Lck_AssertHeld(&oh->mtx);
 	assert(oh->refcnt > 0);
@@ -305,12 +311,11 @@ HSH_Insert(struct worker *wrk, const void *digest, struct objcore *oc,
 	   objecthead. The new object inherits our objhead reference. */
 	oc->objhead = oh;
 	VTAILQ_INSERT_TAIL(&oh->objcs, oc, hsh_list);
-	oc->refcnt++;				// For EXP_Insert
+	EXP_RefNewObjcore(oc);
 	Lck_Unlock(&oh->mtx);
 
 	BAN_RefBan(oc, ban);
 	AN(oc->ban);
-	EXP_Insert(wrk, oc);
 
 	/* Move the object first in the oh list, unbusy it and run the
 	   waitinglist if necessary */
@@ -322,21 +327,25 @@ HSH_Insert(struct worker *wrk, const void *digest, struct objcore *oc,
 		hsh_rush1(wrk, oh, &rush, HSH_RUSH_POLICY);
 	Lck_Unlock(&oh->mtx);
 	hsh_rush2(wrk, &rush);
+
+	EXP_Insert(wrk, oc);
 }
 
 /*---------------------------------------------------------------------
  */
 
 static struct objcore *
-hsh_insert_busyobj(struct worker *wrk, struct objhead *oh)
+hsh_insert_busyobj(const struct worker *wrk, struct objhead *oh)
 {
 	struct objcore *oc;
 
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(wrk->wpriv, WORKER_PRIV_MAGIC);
 	CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
 	Lck_AssertHeld(&oh->mtx);
 
-	oc = wrk->nobjcore;
-	wrk->nobjcore = NULL;
+	oc = wrk->wpriv->nobjcore;
+	wrk->wpriv->nobjcore = NULL;
 	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 
 	AN(oc->flags & OC_F_BUSY);
@@ -360,6 +369,7 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp)
 	vtim_real exp_t_origin;
 	int busy_found;
 	const uint8_t *vary;
+	intmax_t boc_progress;
 	unsigned xid = 0;
 	float dttl = 0.0;
 
@@ -371,6 +381,7 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp)
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
 	wrk = req->wrk;
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(wrk->wpriv, WORKER_PRIV_MAGIC);
 	CHECK_OBJ_NOTNULL(req->http, HTTP_MAGIC);
 	CHECK_OBJ_ORNULL(req->vcf, VCF_MAGIC);
 	AN(hash);
@@ -389,8 +400,8 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp)
 		Lck_Lock(&oh->mtx);
 		req->hash_objhead = NULL;
 	} else {
-		AN(wrk->nobjhead);
-		oh = hash->lookup(wrk, req->digest, &wrk->nobjhead);
+		AN(wrk->wpriv->nobjhead);
+		oh = hash->lookup(wrk, req->digest, &wrk->wpriv->nobjhead);
 	}
 
 	CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
@@ -427,8 +438,11 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp)
 				continue;
 
 			if (oc->boc && oc->boc->vary != NULL &&
-			    !VRY_Match(req, oc->boc->vary))
+			    !req->hash_ignore_vary &&
+			    !VRY_Match(req, oc->boc->vary)) {
+				wrk->strangelove++;
 				continue;
+			}
 
 			busy_found = 1;
 			continue;
@@ -439,15 +453,17 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp)
 
 		if (BAN_CheckObject(wrk, oc, req)) {
 			oc->flags |= OC_F_DYING;
-			EXP_Remove(oc);
+			EXP_Remove(oc, NULL);
 			continue;
 		}
 
-		if (ObjHasAttr(wrk, oc, OA_VARY)) {
+		if (!req->hash_ignore_vary && ObjHasAttr(wrk, oc, OA_VARY)) {
 			vary = ObjGetAttr(wrk, oc, OA_VARY, NULL);
 			AN(vary);
-			if (!VRY_Match(req, vary))
+			if (!VRY_Match(req, vary)) {
+				wrk->strangelove++;
 				continue;
+			}
 		}
 
 		if (req->vcf != NULL) {
@@ -483,7 +499,7 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp)
 		(void)req->vcf->func(req, &oc, &exp_oc, 1);
 
 	if (oc != NULL && oc->flags & OC_F_HFP) {
-		xid = ObjGetXID(wrk, oc);
+		xid = VXID(ObjGetXID(wrk, oc));
 		dttl = EXP_Dttl(req, oc);
 		AN(hsh_deref_objhead_unlock(wrk, &oh, HSH_RUSH_POLICY));
 		wrk->stats->cache_hitpass++;
@@ -495,7 +511,7 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp)
 		*ocp = oc;
 		oc->refcnt++;
 		if (oc->flags & OC_F_HFM) {
-			xid = ObjGetXID(wrk, oc);
+			xid = VXID(ObjGetXID(wrk, oc));
 			dttl = EXP_Dttl(req, oc);
 			*bocp = hsh_insert_busyobj(wrk, oh);
 			Lck_Unlock(&oh->mtx);
@@ -504,7 +520,9 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp)
 			return (HSH_HITMISS);
 		}
 		oc->hits++;
+		boc_progress = oc->boc == NULL ? -1 : oc->boc->fetched_so_far;
 		AN(hsh_deref_objhead_unlock(wrk, &oh, HSH_RUSH_POLICY));
+		Req_LogHit(wrk, req, oc, boc_progress);
 		return (HSH_HIT);
 	}
 
@@ -515,7 +533,7 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp)
 		 * XXX should HFM objects actually have grace/keep ?
 		 * XXX also:  why isn't *ocp = exp_oc ?
 		 */
-		xid = ObjGetXID(wrk, exp_oc);
+		xid = VXID(ObjGetXID(wrk, exp_oc));
 		dttl = EXP_Dttl(req, exp_oc);
 		*bocp = hsh_insert_busyobj(wrk, oh);
 		Lck_Unlock(&oh->mtx);
@@ -524,6 +542,11 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp)
 		return (HSH_HITMISS);
 	}
 
+	if (exp_oc != NULL && exp_oc->boc != NULL)
+		boc_progress = exp_oc->boc->fetched_so_far;
+	else
+		boc_progress = -1;
+
 	if (!busy_found) {
 		*bocp = hsh_insert_busyobj(wrk, oh);
 
@@ -533,6 +556,7 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp)
 			if (EXP_Ttl_grace(req, exp_oc) >= req->t_req) {
 				exp_oc->hits++;
 				Lck_Unlock(&oh->mtx);
+				Req_LogHit(wrk, req, exp_oc, boc_progress);
 				return (HSH_GRACE);
 			}
 		}
@@ -547,6 +571,7 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp)
 		*ocp = exp_oc;
 		exp_oc->hits++;
 		AN(hsh_deref_objhead_unlock(wrk, &oh, 0));
+		Req_LogHit(wrk, req, exp_oc, boc_progress);
 		return (HSH_GRACE);
 	}
 
@@ -634,7 +659,7 @@ hsh_rush2(struct worker *wrk, struct rush *r)
 			 * may be vmod_privs to cleanup and we need a proper
 			 * workerthread for that.
 			 */
-			AZ(Pool_Task(req->sp->pool, &req->task, TASK_QUEUE_RUSH));
+			AZ(Pool_Task(req->sp->pool, req->task, TASK_QUEUE_RUSH));
 		}
 	}
 }
@@ -647,82 +672,109 @@ unsigned
 HSH_Purge(struct worker *wrk, struct objhead *oh, vtim_real ttl_now,
     vtim_dur ttl, vtim_dur grace, vtim_dur keep)
 {
-	struct objcore *oc, **ocp;
-	unsigned spc, ospc, nobj, n, n_tot = 0;
-	int more = 0;
+	struct objcore *oc, *oc_nows[2], **ocp;
+	unsigned i, j, n, n_max, total = 0;
+	int is_purge;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
-	ospc = WS_ReserveAll(wrk->aws);
-	assert(ospc >= sizeof *ocp);
-	/*
-	 * Because of "soft" purges, there might be oc's in the list that has
-	 * the OC_F_PURGED flag set. We do not want to let these slip through,
-	 * so we need to clear the flag before entering the do..while loop.
-	 */
-	Lck_Lock(&oh->mtx);
-	assert(oh->refcnt > 0);
-	VTAILQ_FOREACH(oc, &oh->objcs, hsh_list) {
-		CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
-		assert(oc->objhead == oh);
-		oc->flags &= ~OC_F_PURGED;
-	}
-	Lck_Unlock(&oh->mtx);
 
-	do {
-		more = 0;
-		spc = ospc;
-		nobj = 0;
-		ocp = (void*)wrk->aws->f;
-		Lck_Lock(&oh->mtx);
-		assert(oh->refcnt > 0);
-		VTAILQ_FOREACH(oc, &oh->objcs, hsh_list) {
+	is_purge = (ttl == 0 && grace == 0 && keep == 0);
+	n_max = WS_ReserveLumps(wrk->aws, sizeof *ocp);
+	if (n_max < 2) {
+		/* No space on the workspace. Give it a stack buffer of 2
+		 * elements, which is the minimum for the algorithm
+		 * below. */
+		ocp = oc_nows;
+		n_max = 2;
+	} else
+		ocp = WS_Reservation(wrk->aws);
+	AN(ocp);
+
+	/* Note: This algorithm uses OC references in the list as
+	 * bookmarks, in order to know how far into the list we were when
+	 * releasing the mutex partway through and want to resume
+	 * again. This relies on the list not being reordered while we are
+	 * not holding the mutex. The only place where that happens is in
+	 * HSH_Unbusy(), where an OC_F_BUSY OC is moved first in the
+	 * list. This does not cause problems because we skip OC_F_BUSY
+	 * OCs. */
+
+	Lck_Lock(&oh->mtx);
+	oc = VTAILQ_FIRST(&oh->objcs);
+	n = 0;
+	while (1) {
+		for (; n < n_max && oc != NULL; oc = VTAILQ_NEXT(oc, hsh_list))
+		{
 			CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 			assert(oc->objhead == oh);
 			if (oc->flags & OC_F_BUSY) {
-				/*
-				 * We cannot purge busy objects here, because
+				/* We cannot purge busy objects here, because
 				 * their owners have special rights to them,
 				 * and may nuke them without concern for the
 				 * refcount, which by definition always must
-				 * be one, so they don't check.
-				 */
+				 * be one, so they don't check. */
 				continue;
 			}
 			if (oc->flags & OC_F_DYING)
 				continue;
-			if (oc->flags & OC_F_PURGED) {
-				/*
-				 * We have already called EXP_Rearm on this
-				 * object, and we do not want to do it
-				 * again. Plus the space in the ocp array may
-				 * be limited.
-				 */
-				continue;
-			}
-			if (spc < sizeof *ocp) {
-				/* Iterate if aws is not big enough */
-				more = 1;
-				break;
-			}
+			if (is_purge)
+				oc->flags |= OC_F_DYING;
 			oc->refcnt++;
-			spc -= sizeof *ocp;
-			ocp[nobj++] = oc;
-			oc->flags |= OC_F_PURGED;
+			ocp[n++] = oc;
 		}
+
 		Lck_Unlock(&oh->mtx);
 
-		for (n = 0; n < nobj; n++) {
-			oc = ocp[n];
-			CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
-			EXP_Rearm(oc, ttl_now, ttl, grace, keep);
-			(void)HSH_DerefObjCore(wrk, &oc, 0);
+		if (n == 0) {
+			/* No eligible objcores found. We are finished. */
+			break;
 		}
-		n_tot += nobj;
-	} while (more);
+
+		j = n;
+		if (oc != NULL) {
+			/* There are more objects on the objhead that we
+			 * have not yet looked at, but no more space on
+			 * the objcore reference list. Do not process the
+			 * last one, it will be used as the bookmark into
+			 * the objcore list for the next iteration of the
+			 * outer loop. */
+			j--;
+			assert(j >= 1); /* True because n_max >= 2 */
+		}
+		for (i = 0; i < j; i++) {
+			CHECK_OBJ_NOTNULL(ocp[i], OBJCORE_MAGIC);
+			if (is_purge)
+				EXP_Remove(ocp[i], NULL);
+			else
+				EXP_Rearm(ocp[i], ttl_now, ttl, grace, keep);
+			(void)HSH_DerefObjCore(wrk, &ocp[i], 0);
+			AZ(ocp[i]);
+			total++;
+		}
+
+		if (j == n) {
+			/* No bookmark set, that means we got to the end
+			 * of the objcore list in the previous run and are
+			 * finished. */
+			break;
+		}
+
+		Lck_Lock(&oh->mtx);
+
+		/* Move the bookmark first and continue scanning the
+		 * objcores */
+		CHECK_OBJ_NOTNULL(ocp[j], OBJCORE_MAGIC);
+		ocp[0] = ocp[j];
+		n = 1;
+		oc = VTAILQ_NEXT(ocp[0], hsh_list);
+		CHECK_OBJ_ORNULL(oc, OBJCORE_MAGIC);
+	}
+
 	WS_Release(wrk->aws, 0);
-	Pool_PurgeStat(n_tot);
-	return (n_tot);
+	if (is_purge)
+		Pool_PurgeStat(total);
+	return (total);
 }
 
 /*---------------------------------------------------------------------
@@ -751,11 +803,11 @@ HSH_Fail(struct objcore *oc)
 }
 
 /*---------------------------------------------------------------------
- * Abandon a fetch we will not need
+ * Mark a fetch we will not need as cancelled
  */
 
 static void
-hsh_abandon(struct objcore *oc)
+hsh_cancel(struct objcore *oc)
 {
 	struct objhead *oh;
 
@@ -764,7 +816,7 @@ hsh_abandon(struct objcore *oc)
 	CHECK_OBJ(oh, OBJHEAD_MAGIC);
 
 	Lck_Lock(&oh->mtx);
-	oc->flags |= OC_F_ABANDON;
+	oc->flags |= OC_F_CANCEL;
 	Lck_Unlock(&oh->mtx);
 }
 
@@ -795,7 +847,7 @@ HSH_Cancel(struct worker *wrk, struct objcore *oc, struct boc *boc)
 		AN(oc->flags & OC_F_HFM);
 
 	if (boc != NULL) {
-		hsh_abandon(oc);
+		hsh_cancel(oc);
 		ObjWaitState(oc, BOS_FINISHED);
 	}
 
@@ -817,6 +869,8 @@ HSH_Unbusy(struct worker *wrk, struct objcore *oc)
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+	CHECK_OBJ_NOTNULL(oc->boc, BOC_MAGIC);
+
 	oh = oc->objhead;
 	CHECK_OBJ(oh, OBJHEAD_MAGIC);
 	INIT_OBJ(&rush, RUSH_MAGIC);
@@ -836,7 +890,7 @@ HSH_Unbusy(struct worker *wrk, struct objcore *oc)
 	assert(oh->refcnt > 0);
 	assert(oc->refcnt > 0);
 	if (!(oc->flags & OC_F_PRIVATE))
-		oc->refcnt++;			// For EXP_Insert
+		EXP_RefNewObjcore(oc); /* Takes a ref for expiry */
 	/* XXX: strictly speaking, we should sort in Date: order. */
 	VTAILQ_REMOVE(&oh->objcs, oc, hsh_list);
 	VTAILQ_INSERT_HEAD(&oh->objcs, oc, hsh_list);
@@ -846,8 +900,8 @@ HSH_Unbusy(struct worker *wrk, struct objcore *oc)
 		hsh_rush1(wrk, oh, &rush, HSH_RUSH_POLICY);
 	}
 	Lck_Unlock(&oh->mtx);
-	if (!(oc->flags & OC_F_PRIVATE))
-		EXP_Insert(wrk, oc);
+	EXP_Insert(wrk, oc); /* Does nothing unless EXP_RefNewObjcore was
+			      * called */
 	hsh_rush2(wrk, &rush);
 }
 
@@ -859,15 +913,26 @@ HSH_Unbusy(struct worker *wrk, struct objcore *oc)
 
 void
 HSH_Kill(struct objcore *oc)
+{
+
+	HSH_Replace(oc, NULL);
+}
+
+void
+HSH_Replace(struct objcore *oc, const struct objcore *new_oc)
 {
 
 	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 	CHECK_OBJ_NOTNULL(oc->objhead, OBJHEAD_MAGIC);
+	if (new_oc != NULL) {
+		CHECK_OBJ(new_oc, OBJCORE_MAGIC);
+		assert(oc->objhead == new_oc->objhead);
+	}
 
 	Lck_Lock(&oc->objhead->mtx);
 	oc->flags |= OC_F_DYING;
 	Lck_Unlock(&oc->objhead->mtx);
-	EXP_Remove(oc);
+	EXP_Remove(oc, new_oc);
 }
 
 /*====================================================================
@@ -894,7 +959,7 @@ HSH_Snipe(const struct worker *wrk, struct objcore *oc)
 		Lck_Unlock(&oc->objhead->mtx);
 	}
 	if (retval)
-		EXP_Remove(oc);
+		EXP_Remove(oc, NULL);
 	return (retval);
 }
 
@@ -980,7 +1045,7 @@ HSH_DerefObjCore(struct worker *wrk, struct objcore **ocp, int rushmax)
 	struct objcore *oc;
 	struct objhead *oh;
 	struct rush rush;
-	unsigned r;
+	int r;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	TAKE_OBJ_NOTNULL(oc, ocp, OBJCORE_MAGIC);
