@@ -8,6 +8,8 @@
  * Author: Tollef Fog Heen <tfheen@varnish-software.com>
  * Author: Martin Blix Grydeland <mbgrydeland@varnish-software.com>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -39,13 +41,11 @@
 
 #include "config.h"
 
-#define _WITH_GETLINE
-
 #include <stdlib.h>
 #include <stdio.h>
 #include <unistd.h>
 #include <string.h>
-#include <errno.h>
+#include <signal.h>
 #include <stdarg.h>
 #include <inttypes.h>
 #include <limits.h>
@@ -53,13 +53,15 @@
 #include <time.h>
 #include <math.h>
 
-#include "base64.h"
-#include "vapi/vsm.h"
+#define VOPT_DEFINITION
+#define VOPT_INC "varnishncsa_options.h"
+
+#include "vdef.h"
+
 #include "vapi/vsl.h"
 #include "vapi/voptget.h"
 #include "vas.h"
-#include "vdef.h"
-#include "vcs.h"
+#include "venc.h"
 #include "vsb.h"
 #include "vut.h"
 #include "vqueue.h"
@@ -68,10 +70,9 @@
 #define TIME_FMT "[%d/%b/%Y:%T %z]"
 #define FORMAT "%h %l %u %t \"%r\" %s %b \"%{Referer}i\" \"%{User-agent}i\""
 
-static const char progname[] = "varnishncsa";
+static struct VUT *vut;
 
 struct format;
-struct fragment;
 
 enum e_frag {
 	F_H,			/* %H Proto */
@@ -109,7 +110,7 @@ struct format {
 	char			*string;
 	const char *const	*strptr;
 	char			*time_fmt;
-	int32_t			*int32;
+	int64_t			*int64;
 };
 
 struct watch {
@@ -136,17 +137,18 @@ struct vsl_watch {
 };
 VTAILQ_HEAD(vsl_watch_head, vsl_watch);
 
-struct ctx {
+static struct ctx {
 	/* Options */
 	int			a_opt;
-	int			b_opt;
-	int			c_opt;
 	char			*w_arg;
 
 	FILE			*fo;
 	struct vsb		*vsb;
 	unsigned		gen;
 	VTAILQ_HEAD(,format)	format;
+	int			quote_how;
+	char			*missing_string;
+	char			*missing_int;
 
 	/* State */
 	struct watch_head	watch_vcl_log;
@@ -157,113 +159,63 @@ struct ctx {
 	const char		*hitmiss;
 	const char		*handling;
 	const char		*side;
-	int32_t			vxid;
+	int64_t			vxid;
 } CTX;
 
-static void
-usage(int status)
-{
-	const char **opt;
-
-	fprintf(stderr, "Usage: %s <options>\n\n", progname);
-	fprintf(stderr, "Options:\n");
-	for (opt = vopt_usage; *opt != NULL; opt += 2)
-		fprintf(stderr, " %-25s %s\n", *opt, *(opt + 1));
-	exit(status);
-}
+static void parse_format(const char *format);
 
 static void
 openout(int append)
 {
 
 	AN(CTX.w_arg);
-	CTX.fo = fopen(CTX.w_arg, append ? "a" : "w");
+	if (!strcmp(CTX.w_arg, "-"))
+		CTX.fo = stdout;
+	else
+		CTX.fo = fopen(CTX.w_arg, append ? "a" : "w");
 	if (CTX.fo == NULL)
-		VUT_Error(1, "Can't open output file (%s)",
+		VUT_Error(vut, 1, "Can't open output file (%s)",
 		    strerror(errno));
 }
 
-static int __match_proto__(VUT_cb_f)
-rotateout(void)
+static int v_matchproto_(VUT_cb_f)
+rotateout(struct VUT *v)
 {
 
+	assert(v == vut);
 	AN(CTX.w_arg);
 	AN(CTX.fo);
-	fclose(CTX.fo);
+	(void)fclose(CTX.fo);
 	openout(1);
 	AN(CTX.fo);
 	return (0);
 }
 
-static int __match_proto__(VUT_cb_f)
-flushout(void)
+static int v_matchproto_(VUT_cb_f)
+flushout(struct VUT *v)
 {
 
+	assert(v == vut);
 	AN(CTX.fo);
 	if (fflush(CTX.fo))
 		return (-5);
 	return (0);
 }
 
-static int
-vsb_esc_cat(struct vsb *sb, const char *b, const char *e)
-{
-	AN(b);
-
-	for (; b < e; b++) {
-		if (isspace(*b)) {
-			switch (*b) {
-			case '\n':
-				VSB_cat(sb, "\\n");
-				break;
-			case '\t':
-				VSB_cat(sb, "\\t");
-				break;
-			case '\f':
-				VSB_cat(sb, "\\f");
-				break;
-			case '\r':
-				VSB_cat(sb, "\\r");
-				break;
-			case '\v':
-				VSB_cat(sb, "\\v");
-				break;
-			default:
-				VSB_putc(sb, *b);
-				break;
-			}
-		} else if (isprint(*b)) {
-			switch (*b) {
-			case '"':
-				VSB_cat(sb, "\\\"");
-				break;
-			case '\\':
-				VSB_cat(sb, "\\\\");
-				break;
-			default:
-				VSB_putc(sb, *b);
-				break;
-			}
-		} else
-			VSB_printf(sb, "\\x%02hhx", *b);
-	}
-
-	return (VSB_error(sb));
-}
-
 static inline int
 vsb_fcat(struct vsb *vsb, const struct fragment *f, const char *dflt)
 {
 	if (f->gen == CTX.gen) {
 		assert(f->b <= f->e);
-		return (vsb_esc_cat(vsb, f->b, f->e));
-	}
-	if (dflt)
-		return (vsb_esc_cat(vsb, dflt, dflt + strlen(dflt)));
-	return (-1);
+		VSB_quote(vsb, f->b, f->e - f->b, CTX.quote_how);
+	} else if (dflt)
+		VSB_quote(vsb, dflt, -1, CTX.quote_how);
+	else
+		return (-1);
+	return (0);
 }
 
-static int __match_proto__(format_f)
+static int v_matchproto_(format_f)
 format_string(const struct format *format)
 {
 
@@ -273,7 +225,7 @@ format_string(const struct format *format)
 	return (1);
 }
 
-static int __match_proto__(format_f)
+static int v_matchproto_(format_f)
 format_strptr(const struct format *format)
 {
 
@@ -284,16 +236,16 @@ format_strptr(const struct format *format)
 	return (1);
 }
 
-static int __match_proto__(format_f)
-format_int32(const struct format *format)
+static int v_matchproto_(format_f)
+format_int64(const struct format *format)
 {
 
 	CHECK_OBJ_NOTNULL(format, FORMAT_MAGIC);
-	VSB_printf(CTX.vsb, "%" PRIi32, *format->int32);
+	VSB_printf(CTX.vsb, "%jd", (intmax_t)*format->int64);
 	return (1);
 }
 
-static int __match_proto__(format_f)
+static int v_matchproto_(format_f)
 format_fragment(const struct format *format)
 {
 
@@ -303,21 +255,21 @@ format_fragment(const struct format *format)
 	if (format->frag->gen != CTX.gen) {
 		if (format->string == NULL)
 			return (-1);
-		AZ(vsb_esc_cat(CTX.vsb, format->string,
-		    format->string + strlen(format->string)));
+		VSB_quote(CTX.vsb, format->string, -1, CTX.quote_how);
 		return (0);
 	}
 	AZ(vsb_fcat(CTX.vsb, format->frag, NULL));
 	return (1);
 }
 
-static int __match_proto__(format_f)
+static int v_matchproto_(format_f)
 format_time(const struct format *format)
 {
-	double t_start, t_end;
+	double t_start, t_end, d;
 	char *p;
 	char buf[64];
 	time_t t;
+	intmax_t l;
 	struct tm tm;
 
 	CHECK_OBJ_NOTNULL(format, FORMAT_MAGIC);
@@ -343,29 +295,52 @@ format_time(const struct format *format)
 	} else
 		t_end = t_start;
 
+	AN(format->time_fmt);
+
 	switch (format->time_type) {
-	case 'D':
-		AZ(VSB_printf(CTX.vsb, "%d",
-		    (int)((t_end - t_start) * 1e6)));
-		break;
 	case 't':
-		AN(format->time_fmt);
-		t = t_start;
-		localtime_r(&t, &tm);
-		strftime(buf, sizeof buf, format->time_fmt, &tm);
+		t = (intmax_t)floor(t_start);
+		(void)localtime_r(&t, &tm);
+		AN(strftime(buf, sizeof buf, format->time_fmt, &tm));
 		AZ(VSB_cat(CTX.vsb, buf));
+		return (1);
+	case '3':
+		l = (intmax_t)(modf(t_start, &d) * 1e3);
 		break;
-	case 'T':
-		AZ(VSB_printf(CTX.vsb, "%d", (int)(t_end - t_start)));
+	case '6':
+		l = (intmax_t)(modf(t_start, &d) * 1e6);
+		break;
+	case 'S':
+		l = (intmax_t)t_start;
+		break;
+	case 'M':
+		l = (intmax_t)(t_start * 1e3);
+		break;
+	case 'U':
+		l = (intmax_t)(t_start * 1e6);
+		break;
+	case 's':
+		l = (intmax_t)(t_end - t_start);
+		break;
+	case 'm':
+		l = (intmax_t)((t_end - t_start) * 1e3);
+		break;
+	case 'u':
+		l = (intmax_t)((t_end - t_start) * 1e6);
 		break;
 	default:
 		WRONG("Time format specifier");
 	}
 
+#ifdef __FreeBSD__
+	assert(fmtcheck(format->time_fmt, "%jd") == format->time_fmt);
+#endif
+	AZ(VSB_printf(CTX.vsb, format->time_fmt, l));
+
 	return (1);
 }
 
-static int __match_proto__(format_f)
+static int v_matchproto_(format_f)
 format_requestline(const struct format *format)
 {
 
@@ -385,25 +360,27 @@ format_requestline(const struct format *format)
 	return (1);
 }
 
-static int __match_proto__(format_f)
+static int v_matchproto_(format_f)
 format_auth(const struct format *format)
 {
-	char buf[128];
+	struct vsb *vsb = VSB_new_auto();
+	AN(vsb);
 	char *q;
 
 	if (CTX.frag[F_auth].gen != CTX.gen ||
-	    VB64_decode(buf, sizeof buf, CTX.frag[F_auth].b,
-	    CTX.frag[F_auth].e)) {
+	    VENC_Decode_Base64(vsb, CTX.frag[F_auth].b, CTX.frag[F_auth].e)) {
+		VSB_destroy(&vsb);
 		if (format->string == NULL)
 			return (-1);
-		AZ(vsb_esc_cat(CTX.vsb, format->string,
-		    format->string + strlen(format->string)));
+		VSB_quote(CTX.vsb, format->string, -1, CTX.quote_how);
 		return (0);
 	}
-	q = strchr(buf, ':');
+	AZ(VSB_finish(vsb));
+	q = strchr(VSB_data(vsb), ':');
 	if (q != NULL)
 		*q = '\0';
-	AZ(vsb_esc_cat(CTX.vsb, buf, buf + strlen(buf)));
+	VSB_quote(CTX.vsb, VSB_data(vsb), -1, CTX.quote_how);
+	VSB_destroy(&vsb);
 	return (1);
 }
 
@@ -415,7 +392,9 @@ print(void)
 
 	VSB_clear(CTX.vsb);
 	VTAILQ_FOREACH(f, &CTX.format, list) {
+		CHECK_OBJ_NOTNULL(f, FORMAT_MAGIC);
 		i = (f->func)(f);
+		AZ(VSB_error(CTX.vsb));
 		if (r > i)
 			r = i;
 	}
@@ -437,7 +416,7 @@ addf_string(const char *str)
 	AN(str);
 	ALLOC_OBJ(f, FORMAT_MAGIC);
 	AN(f);
-	f->func = &format_string;
+	f->func = format_string;
 	f->string = strdup(str);
 	AN(f->string);
 	VTAILQ_INSERT_TAIL(&CTX.format, f, list);
@@ -451,7 +430,7 @@ addf_strptr(const char *const *strptr)
 	AN(strptr);
 	ALLOC_OBJ(f, FORMAT_MAGIC);
 	AN(f);
-	f->func = &format_strptr;
+	f->func = format_strptr;
 	f->strptr = strptr;
 	VTAILQ_INSERT_TAIL(&CTX.format, f, list);
 }
@@ -464,7 +443,7 @@ addf_fragment(struct fragment *frag, const char *str)
 	AN(frag);
 	ALLOC_OBJ(f, FORMAT_MAGIC);
 	AN(f);
-	f->func = &format_fragment;
+	f->func = format_fragment;
 	f->frag = frag;
 	if (str != NULL) {
 		f->string = strdup(str);
@@ -474,15 +453,15 @@ addf_fragment(struct fragment *frag, const char *str)
 }
 
 static void
-addf_int32(int32_t *i)
+addf_int64(int64_t *i)
 {
 	struct format *f;
 
 	AN(i);
 	ALLOC_OBJ(f, FORMAT_MAGIC);
 	AN(f);
-	f->func = &format_int32;
-	f->int32 = i;
+	f->func = format_int64;
+	f->int64 = i;
 	VTAILQ_INSERT_TAIL(&CTX.format, f, list);
 }
 
@@ -493,12 +472,48 @@ addf_time(char type, const char *fmt)
 
 	ALLOC_OBJ(f, FORMAT_MAGIC);
 	AN(f);
-	f->func = &format_time;
+	AN(fmt);
+	f->func = format_time;
 	f->time_type = type;
-	if (fmt != NULL) {
-		f->time_fmt = strdup(fmt);
-		AN(f->time_fmt);
+	f->time_fmt = strdup(fmt);
+
+	if (f->time_type == 'T') {
+		if (!strcmp(fmt, "s"))
+			f->time_type = 's';
+		else if (!strcmp(fmt, "ms"))
+			f->time_type = 'm';
+		else if (!strcmp(fmt, "us"))
+			f->time_type = 'u';
+		else
+			VUT_Error(vut, 1, "Unknown specifier: %%{%s}T",
+			    fmt);
+		REPLACE(f->time_fmt, "%jd");
+	} else if (f->time_type == 't') {
+		if (!strcmp(fmt, "sec")) {
+			f->time_type = 'S';
+			REPLACE(f->time_fmt, "%jd");
+		} else if (!strncmp(fmt, "msec", 4)) {
+			fmt += 4;
+			if (!strcmp(fmt, "_frac")) {
+				f->time_type = '3';
+				REPLACE(f->time_fmt, "%03jd");
+			} else if (*fmt == '\0') {
+				f->time_type = 'M';
+				REPLACE(f->time_fmt, "%jd");
+			}
+		} else if (!strncmp(fmt, "usec", 4)) {
+			fmt += 4;
+			if (!strcmp(fmt, "_frac")) {
+				f->time_type = '6';
+				REPLACE(f->time_fmt, "%06jd");
+			} else if (*fmt == '\0') {
+				f->time_type = 'U';
+				REPLACE(f->time_fmt, "%jd");
+			}
+		}
 	}
+
+	AN(f->time_fmt);
 	VTAILQ_INSERT_TAIL(&CTX.format, f, list);
 }
 
@@ -509,7 +524,7 @@ addf_requestline(void)
 
 	ALLOC_OBJ(f, FORMAT_MAGIC);
 	AN(f);
-	f->func = &format_requestline;
+	f->func = format_requestline;
 	VTAILQ_INSERT_TAIL(&CTX.format, f, list);
 }
 
@@ -528,7 +543,7 @@ addf_vcl_log(const char *key)
 
 	ALLOC_OBJ(f, FORMAT_MAGIC);
 	AN(f);
-	f->func = &format_fragment;
+	f->func = format_fragment;
 	f->frag = &w->frag;
 	f->string = strdup("");
 	AN(f->string);
@@ -551,9 +566,9 @@ addf_hdr(struct watch_head *head, const char *key)
 
 	ALLOC_OBJ(f, FORMAT_MAGIC);
 	AN(f);
-	f->func = &format_fragment;
+	f->func = format_fragment;
 	f->frag = &w->frag;
-	f->string = strdup("-");
+	f->string = strdup(CTX.missing_string);
 	AN(f->string);
 	VTAILQ_INSERT_TAIL(&CTX.format, f, list);
 }
@@ -565,6 +580,9 @@ addf_vsl(enum VSL_tag_e tag, long i, const char *prefix)
 
 	ALLOC_OBJ(w, VSL_WATCH_MAGIC);
 	AN(w);
+	if (VSL_tagflags[tag] && CTX.quote_how != VSB_QUOTE_JSON)
+		VUT_Error(vut, 1, "Tag %s can contain control characters",
+		    VSL_tags[tag]);
 	w->tag = tag;
 	assert(i <= INT_MAX);
 	w->idx = i;
@@ -573,7 +591,7 @@ addf_vsl(enum VSL_tag_e tag, long i, const char *prefix)
 		assert(w->prefixlen > 0);
 	}
 	VTAILQ_INSERT_TAIL(&CTX.watch_vsl, w, list);
-	addf_fragment(&w->frag, "-");
+	addf_fragment(&w->frag, CTX.missing_string);
 }
 
 static void
@@ -583,7 +601,7 @@ addf_auth(void)
 
 	ALLOC_OBJ(f, FORMAT_MAGIC);
 	AN(f);
-	f->func = &format_auth;
+	f->func = format_auth;
 	f->string = strdup("-");
 	AN(f->string);
 	VTAILQ_INSERT_TAIL(&CTX.format, f, list);
@@ -593,11 +611,11 @@ static void
 parse_x_format(char *buf)
 {
 	char *e, *r, *s;
+	long lval;
 	int slt;
-	long i;
 
 	if (!strcmp(buf, "Varnish:time_firstbyte")) {
-		addf_fragment(&CTX.frag[F_ttfb], "");
+		addf_fragment(&CTX.frag[F_ttfb], CTX.missing_int);
 		return;
 	}
 	if (!strcmp(buf, "Varnish:hitmiss")) {
@@ -613,7 +631,7 @@ parse_x_format(char *buf)
 		return;
 	}
 	if (!strcmp(buf, "Varnish:vxid")) {
-		addf_int32(&CTX.vxid);
+		addf_int64(&CTX.vxid);
 		return;
 	}
 	if (!strncmp(buf, "VCL_Log:", 8)) {
@@ -626,31 +644,26 @@ parse_x_format(char *buf)
 		while (*e != '\0')
 			e++;
 		if (e == buf)
-			VUT_Error(1, "Missing tag in VSL:");
+			VUT_Error(vut, 1, "Missing tag in VSL:");
 		if (e[-1] == ']') {
 			r = e - 1;
 			while (r > buf && *r != '[')
 				r--;
 			if (r == buf || r[1] == ']')
-				VUT_Error(1, "Syntax error: VSL:%s", buf);
+				VUT_Error(vut, 1, "Syntax error: VSL:%s", buf);
 			e[-1] = '\0';
-			i = strtol(r + 1, &s, 10);
+			lval = strtol(r + 1, &s, 10);
 			if (s != e - 1)
-				VUT_Error(1, "Syntax error: VSL:%s]", buf);
-			if (i <= 0)
-				VUT_Error(1,
+				VUT_Error(vut, 1, "Syntax error: VSL:%s]", buf);
+			if (lval <= 0 || lval > 255) {
+				VUT_Error(vut, 1,
 				    "Syntax error. Field specifier must be"
-				    " positive: %s]",
+				    " between 1 and 255: %s]",
 				    buf);
-			if (i > INT_MAX) {
-				VUT_Error(1,
-				    "Field specifier %ld for the tag VSL:%s]"
-				    " is probably too high",
-				    i, buf);
 			}
 			*r = '\0';
 		} else
-			i = 0;
+			lval = 0;
 		r = buf;
 		while (r < e && *r != ':')
 			r++;
@@ -662,15 +675,19 @@ parse_x_format(char *buf)
 			r = NULL;
 		}
 		if (slt == -2)
-			VUT_Error(1, "Tag not unique: %s", buf);
+			VUT_Error(vut, 1, "Tag not unique: %s", buf);
 		if (slt == -1)
-			VUT_Error(1, "Unknown log tag: %s", buf);
+			VUT_Error(vut, 1, "Unknown log tag: %s", buf);
 		assert(slt >= 0);
 
-		addf_vsl(slt, i, r);
+		addf_vsl((enum VSL_tag_e)slt, lval, r);
 		return;
 	}
-	VUT_Error(1, "Unknown formatting extension: %s", buf);
+	if (!strcmp(buf, "Varnish:default_format")) {
+		parse_format(FORMAT);
+		return;
+	}
+	VUT_Error(vut, 1, "Unknown formatting extension: %s", buf);
 }
 
 static void
@@ -679,6 +696,10 @@ parse_format(const char *format)
 	const char *p, *q;
 	struct vsb *vsb;
 	char buf[256];
+	int b;
+
+	if (format == NULL)
+		format = FORMAT;
 
 	vsb = VSB_new_auto();
 	AN(vsb);
@@ -686,13 +707,12 @@ parse_format(const char *format)
 	for (p = format; *p != '\0'; p++) {
 
 		/* Allow the most essential escape sequences in format */
-		if (*p == '\\') {
-			p++;
-			if (*p == 't')
+		if (*p == '\\' && p[1] != '\0') {
+			if (*++p == 't')
 				AZ(VSB_putc(vsb, '\t'));
 			else if (*p == 'n')
 				AZ(VSB_putc(vsb, '\n'));
-			else if (*p != '\0')
+			else
 				AZ(VSB_putc(vsb, *p));
 			continue;
 		}
@@ -711,28 +731,28 @@ parse_format(const char *format)
 		p++;
 		switch (*p) {
 		case 'b':	/* Body bytes sent */
-			addf_fragment(&CTX.frag[F_b], "-");
+			addf_fragment(&CTX.frag[F_b], CTX.missing_int);
 			break;
 		case 'D':	/* Float request time */
-			addf_time(*p, NULL);
+			addf_time('T', "us");
 			break;
 		case 'h':	/* Client host name / IP Address */
-			addf_fragment(&CTX.frag[F_h], "-");
+			addf_fragment(&CTX.frag[F_h], CTX.missing_string);
 			break;
 		case 'H':	/* Protocol */
 			addf_fragment(&CTX.frag[F_H], "HTTP/1.0");
 			break;
 		case 'I':	/* Bytes received */
-			addf_fragment(&CTX.frag[F_I], "-");
+			addf_fragment(&CTX.frag[F_I], CTX.missing_int);
 			break;
 		case 'l':	/* Client user ID (identd) always '-' */
 			AZ(VSB_putc(vsb, '-'));
 			break;
 		case 'm':	/* Method */
-			addf_fragment(&CTX.frag[F_m], "-");
+			addf_fragment(&CTX.frag[F_m], CTX.missing_string);
 			break;
 		case 'O':	/* Bytes sent */
-			addf_fragment(&CTX.frag[F_O], "-");
+			addf_fragment(&CTX.frag[F_O], CTX.missing_int);
 			break;
 		case 'q':	/* Query string */
 			addf_fragment(&CTX.frag[F_q], "");
@@ -741,29 +761,36 @@ parse_format(const char *format)
 			addf_requestline();
 			break;
 		case 's':	/* Status code */
-			addf_fragment(&CTX.frag[F_s], "-");
+			addf_fragment(&CTX.frag[F_s], CTX.missing_int);
 			break;
 		case 't':	/* strftime */
 			addf_time(*p, TIME_FMT);
 			break;
 		case 'T':	/* Int request time */
-			addf_time(*p, NULL);
+			addf_time(*p, "s");
 			break;
 		case 'u':	/* Remote user from auth */
 			addf_auth();
 			break;
 		case 'U':	/* URL */
-			addf_fragment(&CTX.frag[F_U], "-");
+			addf_fragment(&CTX.frag[F_U], CTX.missing_string);
 			break;
 		case '{':
 			p++;
 			q = p;
-			while (*q && *q != '}')
+			b = 1;
+			while (*q) {
+				if (*q == '{')
+					b++;
+				else if (*q == '}')
+					if (--b == 0)
+						break;
 				q++;
-			if (!*q)
-				VUT_Error(1, "Unmatched bracket at: %s",
+			}
+			if (b > 0)
+				VUT_Error(vut, 1, "Unmatched bracket at: %s",
 				    p - 2);
-			assert(q - p < sizeof buf - 1);
+			assert((unsigned)(q - p) < sizeof buf - 1);
 			strncpy(buf, p, q - p);
 			buf[q - p] = '\0';
 			q++;
@@ -777,18 +804,21 @@ parse_format(const char *format)
 			case 't':
 				addf_time(*q, buf);
 				break;
+			case 'T':
+				addf_time(*q, buf);
+				break;
 			case 'x':
 				parse_x_format(buf);
 				break;
 			default:
-				VUT_Error(1,
+				VUT_Error(vut, 1,
 				    "Unknown format specifier at: %s",
 				    p - 2);
 			}
 			p = q;
 			break;
 		default:
-			VUT_Error(1, "Unknown format specifier at: %s",
+			VUT_Error(vut, 1, "Unknown format specifier at: %s",
 			    p - 1);
 		}
 	}
@@ -800,7 +830,7 @@ parse_format(const char *format)
 		VSB_clear(vsb);
 	}
 
-	VSB_delete(vsb);
+	VSB_destroy(&vsb);
 }
 
 static int
@@ -808,7 +838,7 @@ isprefix(const char *prefix, size_t len, const char *b,
     const char *e, const char **next)
 {
 	assert(len > 0);
-	if (e - b < len || strncasecmp(b, prefix, len))
+	if (e < b + len || strncasecmp(b, prefix, len))
 		return (0);
 	b += len;
 	if (next) {
@@ -831,33 +861,32 @@ frag_fields(int force, const char *b, const char *e, ...)
 	AN(e);
 	va_start(ap, e);
 
-	field = va_arg(ap, int);
-	frag = va_arg(ap, struct fragment *);
-	for (n = 1, q = b; q < e; n++) {
-		/* caller must sort the fields, or this loop will not work: */
-		assert(field >= n);
-		AN(frag);
-
-		p = q;
-		/* Skip WS */
-		while (p < e && isspace(*p))
-			p++;
-		q = p;
-		/* Skip non-WS */
-		while (q < e && !isspace(*q))
-			q++;
-
-		if (field == n) {
-			if (frag->gen != CTX.gen || force) {
-				/* We only grab the same matching field once */
-				frag->gen = CTX.gen;
-				frag->b = p;
-				frag->e = q;
-			}
-			field = va_arg(ap, int);
-			if (field == 0)
-				break;
-			frag = va_arg(ap, struct fragment *);
+	n = 0;
+	while (1) {
+		field = va_arg(ap, int);
+		frag = va_arg(ap, struct fragment *);
+		if (field == 0) {
+			AZ(frag);
+			break;
+		}
+		p = q = NULL;
+		while (n < field) {
+			while (b < e && isspace(*b))
+				b++;
+			p = b;
+			while (b < e && !isspace(*b))
+				b++;
+			q = b;
+			n++;
+		}
+		assert(p != NULL && q != NULL);
+		if (p >= e || q <= p)
+			continue;
+		if (frag->gen != CTX.gen || force) {
+			/* We only grab the same matching field once */
+			frag->gen = CTX.gen;
+			frag->b = p;
+			frag->e = q;
 		}
 	}
 	va_end(ap);
@@ -891,6 +920,7 @@ process_hdr(const struct watch_head *head, const char *b, const char *e)
 	const char *p;
 
 	VTAILQ_FOREACH(w, head, list) {
+		CHECK_OBJ_NOTNULL(w, WATCH_MAGIC);
 		if (!isprefix(w->key, w->keylen, b, e, &p))
 			continue;
 		frag_line(1, p, e, &w->frag);
@@ -903,7 +933,6 @@ process_vsl(const struct vsl_watch_head *head, enum VSL_tag_e tag,
 {
 	struct vsl_watch *w;
 	const char *p;
-
 	VTAILQ_FOREACH(w, head, list) {
 		CHECK_OBJ_NOTNULL(w, VSL_WATCH_MAGIC);
 		if (tag != w->tag)
@@ -919,51 +948,47 @@ process_vsl(const struct vsl_watch_head *head, enum VSL_tag_e tag,
 	}
 }
 
-static int __match_proto__(VSLQ_dispatch_f)
+static int v_matchproto_(VSLQ_dispatch_f)
 dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
     void *priv)
 {
 	struct VSL_transaction *t;
-	unsigned tag;
+	enum VSL_tag_e tag;
 	const char *b, *e, *p;
 	struct watch *w;
-	int i, skip, be_mark;
+	int i, skip;
 
 	(void)vsl;
 	(void)priv;
 
-#define BACKEND_MARKER (INT_MAX / 2 + 1)
-	assert(BACKEND_MARKER >= VSL_t__MAX);
-
 	for (t = pt[0]; t != NULL; t = *++pt) {
 		CTX.gen++;
 
-		/* Consider client requests only if in client mode.
-		   Consider backend requests only if in backend mode. */
-		if (t->type == VSL_t_req && CTX.c_opt) {
+		if (t->type == VSL_t_req) {
 			CTX.side = "c";
-			be_mark = 0;
-		} else if (t->type == VSL_t_bereq && CTX.b_opt) {
+		} else if (t->type == VSL_t_bereq) {
 			CTX.side = "b";
-			be_mark = BACKEND_MARKER;
 		} else
 			continue;
-		if (t->reason == VSL_r_esi)
-			/* Skip ESI requests */
-			continue;
+
 		CTX.hitmiss = "-";
 		CTX.handling = "-";
 		CTX.vxid = t->vxid;
 		skip = 0;
 		while (skip == 0 && 1 == VSL_Next(t->c)) {
 			tag = VSL_TAG(t->c->rec.ptr);
+			if (VSL_tagflags[tag] &&
+			    CTX.quote_how != VSB_QUOTE_JSON)
+				continue;
+
 			b = VSL_CDATA(t->c->rec.ptr);
 			e = b + VSL_LEN(t->c->rec.ptr);
-			while (e > b && e[-1] == '\0')
-				e--;
+			if (!VSL_tagflags[tag]) {
+				while (e > b && e[-1] == '\0')
+					e--;
+			}
 
-			switch (tag + be_mark) {
-			case SLT_HttpGarbage + BACKEND_MARKER:
+			switch (tag) {
 			case SLT_HttpGarbage:
 				skip = 1;
 				break;
@@ -973,9 +998,9 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 				    4, &CTX.frag[F_O],
 				    0, NULL);
 				break;
-			case (SLT_BackendStart + BACKEND_MARKER):
+			case SLT_BackendOpen:
 				frag_fields(1, b, e,
-				    1, &CTX.frag[F_h],
+				    3, &CTX.frag[F_h],
 				    0, NULL);
 				break;
 			case SLT_ReqStart:
@@ -983,11 +1008,11 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 				    1, &CTX.frag[F_h],
 				    0, NULL);
 				break;
-			case (SLT_BereqMethod + BACKEND_MARKER):
+			case SLT_BereqMethod:
 			case SLT_ReqMethod:
 				frag_line(0, b, e, &CTX.frag[F_m]);
 				break;
-			case (SLT_BereqURL + BACKEND_MARKER):
+			case SLT_BereqURL:
 			case SLT_ReqURL:
 				p = memchr(b, '?', e - b);
 				if (p == NULL)
@@ -995,15 +1020,15 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 				frag_line(0, b, p, &CTX.frag[F_U]);
 				frag_line(0, p, e, &CTX.frag[F_q]);
 				break;
-			case (SLT_BereqProtocol + BACKEND_MARKER):
+			case SLT_BereqProtocol:
 			case SLT_ReqProtocol:
 				frag_line(0, b, e, &CTX.frag[F_H]);
 				break;
-			case (SLT_BerespStatus + BACKEND_MARKER):
+			case SLT_BerespStatus:
 			case SLT_RespStatus:
 				frag_line(1, b, e, &CTX.frag[F_s]);
 				break;
-			case (SLT_BereqAcct + BACKEND_MARKER):
+			case SLT_BereqAcct:
 			case SLT_ReqAcct:
 				frag_fields(0, b, e,
 				    3, &CTX.frag[F_I],
@@ -1011,7 +1036,6 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 				    6, &CTX.frag[F_O],
 				    0, NULL);
 				break;
-			case (SLT_Timestamp + BACKEND_MARKER):
 			case SLT_Timestamp:
 #define ISPREFIX(a, b, c, d)	isprefix(a, strlen(a), b, c, d)
 				if (ISPREFIX("Start:", b, e, &p)) {
@@ -1031,8 +1055,9 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 					    &CTX.frag[F_ttfb], 0, NULL);
 				}
 				break;
-			case (SLT_BereqHeader + BACKEND_MARKER):
+			case SLT_BereqHeader:
 			case SLT_ReqHeader:
+				process_hdr(&CTX.watch_reqhdr, b, e);
 				if (ISPREFIX("Authorization:", b, e, &p) &&
 				    ISPREFIX("basic ", p, e, &p))
 					frag_line(0, p, e,
@@ -1042,6 +1067,10 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 					    &CTX.frag[F_host]);
 #undef ISPREFIX
 				break;
+			case SLT_BerespHeader:
+			case SLT_RespHeader:
+				process_hdr(&CTX.watch_resphdr, b, e);
+				break;
 			case SLT_VCL_call:
 				if (!strcasecmp(b, "recv")) {
 					CTX.hitmiss = "-";
@@ -1070,7 +1099,6 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 				} else if (!strcasecmp(b, "restart"))
 					skip = 1;
 				break;
-			case (SLT_VCL_Log + BACKEND_MARKER):
 			case SLT_VCL_Log:
 				VTAILQ_FOREACH(w, &CTX.watch_vcl_log, list) {
 					CHECK_OBJ_NOTNULL(w, WATCH_MAGIC);
@@ -1085,13 +1113,6 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 				break;
 			}
 
-			if ((tag == SLT_ReqHeader && CTX.c_opt) ||
-			    (tag == SLT_BereqHeader && CTX.b_opt))
-				process_hdr(&CTX.watch_reqhdr, b, e);
-			else if ((tag == SLT_RespHeader && CTX.c_opt) ||
-			    (tag == SLT_BerespHeader && CTX.b_opt))
-				process_hdr(&CTX.watch_resphdr, b, e);
-
 			process_vsl(&CTX.watch_vsl, tag, b, e);
 		}
 		if (skip)
@@ -1104,12 +1125,6 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 	return (0);
 }
 
-static int __match_proto__(VUT_cb_f)
-sighup(void)
-{
-	return (1);
-}
-
 static char *
 read_format(const char *formatfile)
 {
@@ -1120,18 +1135,18 @@ read_format(const char *formatfile)
 
 	fmtfile = fopen(formatfile, "r");
 	if (fmtfile == NULL)
-		VUT_Error(1, "Can't open format file (%s)",
+		VUT_Error(vut, 1, "Can't open format file (%s)",
 		    strerror(errno));
+	AN(fmtfile);
 	fmtlen = getline(&fmt, &len, fmtfile);
 	if (fmtlen == -1) {
 		free(fmt);
 		if (feof(fmtfile))
-			VUT_Error(1, "Empty format file");
-		else
-			VUT_Error(1, "Can't read format from file (%s)",
-			    strerror(errno));
+			VUT_Error(vut, 1, "Empty format file");
+		VUT_Error(vut, 1, "Can't read format from file (%s)",
+		    strerror(errno));
 	}
-	fclose(fmtfile);
+	AZ(fclose(fmtfile));
 	if (fmt[fmtlen - 1] == '\n')
 		fmt[fmtlen - 1] = '\0';
 	return (fmt);
@@ -1142,7 +1157,10 @@ main(int argc, char * const *argv)
 {
 	signed char opt;
 	char *format = NULL;
+	int mode_opt = 0;
 
+	vut = VUT_InitProg(argc, argv, &vopt_spec);
+	AN(vut);
 	memset(&CTX, 0, sizeof CTX);
 	VTAILQ_INIT(&CTX.format);
 	VTAILQ_INIT(&CTX.watch_vcl_log);
@@ -1151,88 +1169,93 @@ main(int argc, char * const *argv)
 	VTAILQ_INIT(&CTX.watch_vsl);
 	CTX.vsb = VSB_new_auto();
 	AN(CTX.vsb);
-	VB64_init();
-	VUT_Init(progname);
+	CTX.quote_how = VSB_QUOTE_ESCHEX;
+	REPLACE(CTX.missing_string, "-");
+	REPLACE(CTX.missing_int, "-");
 
-	while ((opt = getopt(argc, argv, vopt_optstring)) != -1) {
+	tzset();		// We use localtime_r(3)
+
+	while ((opt = getopt(argc, argv, vopt_spec.vopt_optstring)) != -1) {
 		switch (opt) {
 		case 'a':
 			/* Append to file */
 			CTX.a_opt = 1;
 			break;
-		case 'b':
-			/* backend mode */
-			CTX.b_opt = 1;
-			break;
-		case 'c':
-			/* client mode */
-			CTX.c_opt = 1;
+		case 'b': /* backend mode */
+		case 'c': /* client mode */
+		case 'E': /* show ESI */
+			AN(VUT_Arg(vut, opt, NULL));
+			mode_opt = 1;
 			break;
 		case 'F':
-			/* Format string */
 			if (format != NULL)
-				free(format);
-			format = strdup(optarg);
-			AN(format);
+				VUT_Error(vut, 1, "Format already set");
+			REPLACE(format, optarg);
 			break;
 		case 'f':
-			/* Format string from file */
 			if (format != NULL)
-				free(format);
+				VUT_Error(vut, 1, "Format already set");
+			/* Format string from file */
 			format = read_format(optarg);
 			AN(format);
 			break;
 		case 'h':
 			/* Usage help */
-			usage(0);
+			VUT_Usage(vut, &vopt_spec, 0);
+			break;
+		case 'j':
+			REPLACE(CTX.missing_string, "");
+			REPLACE(CTX.missing_int, "0");
+			CTX.quote_how = VSB_QUOTE_JSON;
+			break;
 		case 'w':
 			/* Write to file */
 			REPLACE(CTX.w_arg, optarg);
 			break;
 		default:
-			if (!VUT_Arg(opt, optarg))
-				usage(1);
+			if (!VUT_Arg(vut, opt, optarg))
+				VUT_Usage(vut, &vopt_spec, 1);
 		}
 	}
+
 	/* default is client mode: */
-	if (!CTX.b_opt)
-		CTX.c_opt = 1;
+	if (!mode_opt)
+		AN(VUT_Arg(vut, 'c', NULL));
 
 	if (optind != argc)
-		usage(1);
+		VUT_Usage(vut, &vopt_spec, 1);
 
-	if (VUT.D_opt && !CTX.w_arg)
-		VUT_Error(1, "Missing -w option");
+	if (vut->D_opt && !CTX.w_arg)
+		VUT_Error(vut, 1, "Missing -w option");
+
+	if (vut->D_opt && !strcmp(CTX.w_arg, "-"))
+		VUT_Error(vut, 1, "Daemon cannot write to stdout");
 
 	/* Check for valid grouping mode */
-	assert(VUT.g_arg < VSL_g__MAX);
-	if (VUT.g_arg != VSL_g_vxid && VUT.g_arg != VSL_g_request)
-		VUT_Error(1, "Invalid grouping mode: %s",
-		    VSLQ_grouping[VUT.g_arg]);
+	assert(vut->g_arg < VSL_g__MAX);
+	if (vut->g_arg != VSL_g_vxid && vut->g_arg != VSL_g_request)
+		VUT_Error(vut, 1, "Invalid grouping mode: %s",
+		    VSLQ_grouping[vut->g_arg]);
 
 	/* Prepare output format */
-	if (format == NULL)
-		format = strdup(FORMAT);
 	parse_format(format);
-	free(format);
-	format = NULL;
+	REPLACE(format, NULL);
 
 	/* Setup output */
-	VUT.dispatch_f = &dispatch_f;
-	VUT.dispatch_priv = NULL;
-	VUT.sighup_f = sighup;
+	vut->dispatch_f = dispatch_f;
+	vut->dispatch_priv = NULL;
 	if (CTX.w_arg) {
 		openout(CTX.a_opt);
 		AN(CTX.fo);
-		if (VUT.D_opt)
-			VUT.sighup_f = &rotateout;
+		if (vut->D_opt)
+			vut->sighup_f = rotateout;
 	} else
 		CTX.fo = stdout;
-	VUT.idle_f = &flushout;
+	vut->idle_f = flushout;
 
-	VUT_Setup();
-	VUT_Main();
-	VUT_Fini();
+	VUT_Setup(vut);
+	(void)VUT_Main(vut);
+	VUT_Fini(&vut);
 
 	exit(0);
 }
