@@ -39,6 +39,7 @@
 #include <string.h>
 #include <unistd.h>
 
+#include "vsa.h"
 #include "vtc.h"
 
 #include "vtcp.h"
@@ -50,19 +51,18 @@ struct server {
 	char			*name;
 	struct vtclog		*vl;
 	VTAILQ_ENTRY(server)	list;
+	struct vtc_sess		*vsp;
 	char			run;
 
-	int			repeat;
-	unsigned		keepalive;
 	char			*spec;
 
 	int			depth;
 	int			sock;
 	int			fd;
-	int			rcvbuf;
+	unsigned		is_dispatch;
 	char			listen[256];
-	char			aaddr[32];
-	char			aport[32];
+	char			aaddr[VTCP_ADDRBUFSIZE];
+	char			aport[VTCP_PORTBUFSIZE];
 
 	pthread_t		tp;
 };
@@ -85,17 +85,18 @@ server_new(const char *name, struct vtclog *vl)
 	ALLOC_OBJ(s, SERVER_MAGIC);
 	AN(s);
 	REPLACE(s->name, name);
-	s->vl = vtc_logopen(s->name);
+	s->vl = vtc_logopen("%s", s->name);
 	AN(s->vl);
+	s->vsp = Sess_New(s->vl, name);
+	AN(s->vsp);
 
-	bprintf(s->listen, "%s", "127.0.0.1 0");
-	s->repeat = 1;
+	bprintf(s->listen, "%s", default_listen_addr);
 	s->depth = 10;
 	s->sock = -1;
 	s->fd = -1;
-	AZ(pthread_mutex_lock(&server_mtx));
+	PTOK(pthread_mutex_lock(&server_mtx));
 	VTAILQ_INSERT_TAIL(&servers, s, list);
-	AZ(pthread_mutex_unlock(&server_mtx));
+	PTOK(pthread_mutex_unlock(&server_mtx));
 	return (s);
 }
 
@@ -108,6 +109,7 @@ server_delete(struct server *s)
 {
 
 	CHECK_OBJ_NOTNULL(s, SERVER_MAGIC);
+	Sess_Destroy(&s->vsp);
 	macro_undef(s->vl, s->name, "addr");
 	macro_undef(s->vl, s->name, "port");
 	macro_undef(s->vl, s->name, "sock");
@@ -164,7 +166,7 @@ server_listen_uds(struct server *s, const char **errp)
 	errno = 0;
 	if (unlink(s->listen) != 0 && errno != ENOENT)
 		vtc_fatal(s->vl, "Could not unlink %s before bind: %s",
-			  s->listen, strerror(errno));
+		    s->listen, strerror(errno));
 	/*
 	 * Temporarily set the umask to 0 to avoid issues with
 	 * permissions.
@@ -183,17 +185,27 @@ server_listen_uds(struct server *s, const char **errp)
 static void
 server_listen_tcp(struct server *s, const char **errp)
 {
+	char buf[vsa_suckaddr_len];
+	const struct suckaddr *sua;
+
 	s->sock = VTCP_listen_on(s->listen, "0", s->depth, errp);
 	if (*errp != NULL)
 		return;
 	assert(s->sock > 0);
-	VTCP_myname(s->sock, s->aaddr, sizeof s->aaddr,
+	sua = VSA_getsockname(s->sock, buf, sizeof buf);
+	AN(sua);
+	VTCP_name(sua, s->aaddr, sizeof s->aaddr,
 	    s->aport, sizeof s->aport);
+
+	/* Record the actual port, and reuse it on subsequent starts */
+	if (VSA_Get_Proto(sua) == AF_INET)
+		bprintf(s->listen, "%s:%s", s->aaddr, s->aport);
+	else
+		bprintf(s->listen, "[%s]:%s", s->aaddr, s->aport);
+
 	macro_def(s->vl, s->name, "addr", "%s", s->aaddr);
 	macro_def(s->vl, s->name, "port", "%s", s->aport);
-	macro_def(s->vl, s->name, "sock", "%s %s", s->aaddr, s->aport);
-	/* Record the actual port, and reuse it on subsequent starts */
-	bprintf(s->listen, "%s %s", s->aaddr, s->aport);
+	macro_def(s->vl, s->name, "sock", "%s", s->listen);
 }
 
 static void
@@ -205,10 +217,10 @@ server_listen(struct server *s)
 
 	if (s->sock >= 0)
 		VTCP_close(&s->sock);
-	if (*s->listen != '/')
-		server_listen_tcp(s, &err);
-	else
+	if (VUS_is(s->listen))
 		server_listen_uds(s, &err);
+	else
+		server_listen_tcp(s, &err);
 	if (err != NULL)
 		vtc_fatal(s->vl,
 		    "Server listen address (%s) cannot be resolved: %s",
@@ -219,56 +231,62 @@ server_listen(struct server *s)
  * Server thread
  */
 
-static void *
-server_thread(void *priv)
+static int
+server_conn(void *priv, struct vtclog *vl)
 {
 	struct server *s;
-	struct vtclog *vl;
-	int i, j, fd;
 	struct sockaddr_storage addr_s;
 	struct sockaddr *addr;
-	socklen_t l;
 	char abuf[VTCP_ADDRBUFSIZE];
 	char pbuf[VTCP_PORTBUFSIZE];
+	socklen_t l;
+	int fd;
 
 	CAST_OBJ_NOTNULL(s, priv, SERVER_MAGIC);
-	assert(s->sock >= 0);
 
-	vl = vtc_logopen(s->name);
-	pthread_cleanup_push(vtc_logclose, vl);
-
-	vtc_log(vl, 2, "Started on %s (%u iterations%s)", s->listen,
-		s->repeat, s->keepalive ? " using keepalive" : "");
-	for (i = 0; i < s->repeat; i++) {
-		addr = (void*)&addr_s;
-		l = sizeof addr_s;
-		fd = accept(s->sock, addr, &l);
-		if (fd < 0)
-			vtc_fatal(vl, "Accept failed: %s", strerror(errno));
-		if (*s->listen != '/') {
-			VTCP_hisname(fd, abuf, sizeof abuf, pbuf, sizeof pbuf);
-			vtc_log(vl, 3, "accepted fd %d %s %s", fd, abuf, pbuf);
-		} else
-			vtc_log(vl, 3, "accepted fd %d 0.0.0.0 0", fd);
-		if (! s->keepalive)
-			fd = http_process(vl, s->spec, fd, &s->sock, s->listen,
-			    s->rcvbuf);
-		else
-			while (fd >= 0 && i++ < s->repeat)
-				fd = http_process(vl, s->spec, fd,
-				    &s->sock, s->listen, s->rcvbuf);
-		vtc_log(vl, 3, "shutting fd %d", fd);
-		j = shutdown(fd, SHUT_WR);
-		if (!VTCP_Check(j))
-			vtc_fatal(vl, "Shutdown failed: %s", strerror(errno));
-		VTCP_close(&fd);
+	addr = (void*)&addr_s;
+	l = sizeof addr_s;
+	fd = accept(s->sock, addr, &l);
+	if (fd < 0)
+		vtc_fatal(vl, "Accept failed: %s", strerror(errno));
+	if (VUS_is(s->listen))
+		vtc_log(vl, 3, "accepted fd %d 0.0.0.0 0", fd);
+	else {
+		VTCP_hisname(fd, abuf, sizeof abuf, pbuf, sizeof pbuf);
+		vtc_log(vl, 3, "accepted fd %d %s %s", fd, abuf, pbuf);
 	}
-	vtc_log(vl, 2, "Ending");
-	pthread_cleanup_pop(0);
-	vtc_logclose(vl);
-	return (NULL);
+	return (fd);
 }
 
+static void
+server_disc(void *priv, struct vtclog *vl, int *fdp)
+{
+	int j;
+	struct server *s;
+
+	CAST_OBJ_NOTNULL(s, priv, SERVER_MAGIC);
+	vtc_log(vl, 3, "shutting fd %d", *fdp);
+	j = shutdown(*fdp, SHUT_WR);
+	if (!vtc_stop && !VTCP_Check(j))
+		vtc_fatal(vl, "Shutdown failed: %s", strerror(errno));
+	VTCP_close(fdp);
+}
+
+static void
+server_start_thread(struct server *s)
+{
+
+	s->run = 1;
+	s->tp = Sess_Start_Thread(
+	    s,
+	    s->vsp,
+	    server_conn,
+	    server_disc,
+	    s->listen,
+	    &s->sock,
+	    s->spec
+	);
+}
 
 /**********************************************************************
  * Start the server thread
@@ -281,8 +299,7 @@ server_start(struct server *s)
 	vtc_log(s->vl, 2, "Starting server");
 	server_listen(s);
 	vtc_log(s->vl, 1, "Listen on %s", s->listen);
-	s->run = 1;
-	AZ(pthread_create(&s->tp, NULL, server_thread, s));
+	server_start_thread(s);
 }
 
 /**********************************************************************
@@ -298,13 +315,13 @@ server_dispatch_wrk(void *priv)
 	CAST_OBJ_NOTNULL(s, priv, SERVER_MAGIC);
 	assert(s->sock < 0);
 
-	vl = vtc_logopen(s->name);
+	vl = vtc_logopen("%s", s->name);
 	pthread_cleanup_push(vtc_logclose, vl);
 
 	fd = s->fd;
 
 	vtc_log(vl, 3, "start with fd %d", fd);
-	fd = http_process(vl, s->spec, fd, &s->sock, s->listen, s->rcvbuf);
+	fd = sess_process(vl, s->vsp, s->spec, fd, &s->sock, s->listen);
 	vtc_log(vl, 3, "shutting fd %d", fd);
 	j = shutdown(fd, SHUT_WR);
 	if (!VTCP_Check(j))
@@ -320,7 +337,8 @@ static void *
 server_dispatch_thread(void *priv)
 {
 	struct server *s, *s2;
-	int sn = 1, fd;
+	static int sn = 1;
+	int fd;
 	char snbuf[8];
 	struct vtclog *vl;
 	struct sockaddr_storage addr_s;
@@ -330,7 +348,7 @@ server_dispatch_thread(void *priv)
 	CAST_OBJ_NOTNULL(s, priv, SERVER_MAGIC);
 	assert(s->sock >= 0);
 
-	vl = vtc_logopen(s->name);
+	vl = vtc_logopen("%s", s->name);
 	pthread_cleanup_push(vtc_logclose, vl);
 
 	vtc_log(vl, 2, "Dispatch started on %s", s->listen);
@@ -344,11 +362,12 @@ server_dispatch_thread(void *priv)
 		bprintf(snbuf, "s%d", sn++);
 		vtc_log(vl, 3, "dispatch fd %d -> %s", fd, snbuf);
 		s2 = server_new(snbuf, vl);
+		s2->is_dispatch = 1;
 		s2->spec = s->spec;
 		bstrcpy(s2->listen, s->listen);
 		s2->fd = fd;
 		s2->run = 1;
-		AZ(pthread_create(&s2->tp, NULL, server_dispatch_wrk, s2));
+		PTOK(pthread_create(&s2->tp, NULL, server_dispatch_wrk, s2));
 	}
 	pthread_cleanup_pop(0);
 	vtc_logclose(vl);
@@ -362,7 +381,7 @@ server_dispatch(struct server *s)
 	server_listen(s);
 	vtc_log(s->vl, 2, "Starting dispatch server");
 	s->run = 1;
-	AZ(pthread_create(&s->tp, NULL, server_dispatch_thread, s));
+	PTOK(pthread_create(&s->tp, NULL, server_dispatch_thread, s));
 }
 
 /**********************************************************************
@@ -377,7 +396,7 @@ server_break(struct server *s)
 	CHECK_OBJ_NOTNULL(s, SERVER_MAGIC);
 	vtc_log(s->vl, 2, "Breaking for server");
 	(void)pthread_cancel(s->tp);
-	AZ(pthread_join(s->tp, &res));
+	PTOK(pthread_join(s->tp, &res));
 	VTCP_close(&s->sock);
 	s->tp = 0;
 	s->run = 0;
@@ -394,7 +413,7 @@ server_wait(struct server *s)
 
 	CHECK_OBJ_NOTNULL(s, SERVER_MAGIC);
 	vtc_log(s->vl, 2, "Waiting for server (%d/%d)", s->sock, s->fd);
-	AZ(pthread_join(s->tp, &res));
+	PTOK(pthread_join(s->tp, &res));
 	if (res != NULL && !vtc_stop)
 		vtc_fatal(s->vl, "Server returned \"%p\"",
 		    (char *)res);
@@ -411,19 +430,21 @@ cmd_server_gen_vcl(struct vsb *vsb)
 {
 	struct server *s;
 
-	AZ(pthread_mutex_lock(&server_mtx));
+	PTOK(pthread_mutex_lock(&server_mtx));
 	VTAILQ_FOREACH(s, &servers, list) {
-		if (*s->listen != '/')
+		if (s->is_dispatch)
+			continue;
+
+		if (VUS_is(s->listen))
 			VSB_printf(vsb,
-				   "backend %s { .host = \"%s\"; "
-				   ".port = \"%s\"; }\n",
-				   s->name, s->aaddr, s->aport);
+			   "backend %s { .path = \"%s\"; }\n",
+			   s->name, s->listen);
 		else
 			VSB_printf(vsb,
-				   "backend %s { .path = \"%s\"; }\n",
-				   s->name, s->listen);
+			   "backend %s { .host = \"%s\"; .port = \"%s\"; }\n",
+			   s->name, s->aaddr, s->aport);
 	}
-	AZ(pthread_mutex_unlock(&server_mtx));
+	PTOK(pthread_mutex_unlock(&server_mtx));
 }
 
 
@@ -436,9 +457,9 @@ cmd_server_gen_haproxy_conf(struct vsb *vsb)
 {
 	struct server *s;
 
-	AZ(pthread_mutex_lock(&server_mtx));
+	PTOK(pthread_mutex_lock(&server_mtx));
 	VTAILQ_FOREACH(s, &servers, list) {
-		if (*s->listen != '/')
+		if (! VUS_is(s->listen))
 			VSB_printf(vsb,
 			   "\n    backend be%s\n"
 			   "\tserver srv%s %s:%s\n",
@@ -447,7 +468,7 @@ cmd_server_gen_haproxy_conf(struct vsb *vsb)
 			INCOMPL();
 	}
 	VTAILQ_FOREACH(s, &servers, list) {
-		if (*s->listen != '/')
+		if (! VUS_is(s->listen))
 			VSB_printf(vsb,
 			   "\n    frontend http%s\n"
 			   "\tuse_backend be%s\n"
@@ -456,7 +477,7 @@ cmd_server_gen_haproxy_conf(struct vsb *vsb)
 		else
 			INCOMPL();
 	}
-	AZ(pthread_mutex_unlock(&server_mtx));
+	PTOK(pthread_mutex_unlock(&server_mtx));
 }
 
 
@@ -470,17 +491,16 @@ cmd_server(CMD_ARGS)
 	struct server *s;
 
 	(void)priv;
-	(void)cmd;
 
 	if (av == NULL) {
 		/* Reset and free */
 		while (1) {
-			AZ(pthread_mutex_lock(&server_mtx));
+			PTOK(pthread_mutex_lock(&server_mtx));
 			s = VTAILQ_FIRST(&servers);
 			CHECK_OBJ_ORNULL(s, SERVER_MAGIC);
 			if (s != NULL)
 				VTAILQ_REMOVE(&servers, s, list);
-			AZ(pthread_mutex_unlock(&server_mtx));
+			PTOK(pthread_mutex_unlock(&server_mtx));
 			if (s == NULL)
 				break;
 			if (s->run) {
@@ -497,11 +517,11 @@ cmd_server(CMD_ARGS)
 	AZ(strcmp(av[0], "server"));
 	av++;
 
-	AZ(pthread_mutex_lock(&server_mtx));
+	PTOK(pthread_mutex_lock(&server_mtx));
 	VTAILQ_FOREACH(s, &servers, list)
 		if (!strcmp(s->name, av[0]))
 			break;
-	AZ(pthread_mutex_unlock(&server_mtx));
+	PTOK(pthread_mutex_unlock(&server_mtx));
 	if (s == NULL)
 		s = server_new(av[0], vl);
 	CHECK_OBJ_NOTNULL(s, SERVER_MAGIC);
@@ -530,15 +550,10 @@ cmd_server(CMD_ARGS)
 			server_wait(s);
 
 		AZ(s->run);
-		if (!strcmp(*av, "-repeat")) {
-			s->repeat = atoi(av[1]);
-			av++;
+
+		if (Sess_GetOpt(s->vsp, &av))
 			continue;
-		}
-		if (!strcmp(*av, "-keepalive")) {
-			s->keepalive = 1;
-			continue;
-		}
+
 		if (!strcmp(*av, "-listen")) {
 			if (s->sock >= 0)
 				VTCP_close(&s->sock);
@@ -546,11 +561,6 @@ cmd_server(CMD_ARGS)
 			av++;
 			continue;
 		}
-		if (!strcmp(*av, "-rcvbuf")) {
-			s->rcvbuf = atoi(av[1]);
-			av++;
-			continue;
-		}
 		if (!strcmp(*av, "-start")) {
 			server_start(s);
 			continue;
@@ -571,5 +581,5 @@ cmd_server(CMD_ARGS)
 void
 init_server(void)
 {
-	AZ(pthread_mutex_init(&server_mtx, NULL));
+	PTOK(pthread_mutex_init(&server_mtx, NULL));
 }
