@@ -163,8 +163,6 @@ void VCA_Shutdown(void);
 /* cache_backend_cfg.c */
 void VBE_InitCfg(void);
 
-void VBP_Init(void);
-
 /* cache_ban.c */
 
 /* for stevedoes resurrecting bans */
@@ -189,7 +187,8 @@ void VDI_Event(const struct director *d, enum vcl_event_e ev);
 void VDI_Init(void);
 
 /* cache_deliver_proc.c */
-void VDP_Init(struct vdp_ctx *vdx, struct worker *wrk, struct vsl_log *vsl,
+void VDP_Fini(const struct vdp_ctx *vdc);
+void VDP_Init(struct vdp_ctx *vdc, struct worker *wrk, struct vsl_log *vsl,
     struct req *req);
 uint64_t VDP_Close(struct vdp_ctx *, struct objcore *, struct boc *);
 void VDP_Panic(struct vsb *vsb, const struct vdp_ctx *vdc);
@@ -206,7 +205,7 @@ vtim_real EXP_Ttl(const struct req *, const struct objcore *);
 vtim_real EXP_Ttl_grace(const struct req *, const struct objcore *oc);
 void EXP_RefNewObjcore(struct objcore *);
 void EXP_Insert(struct worker *wrk, struct objcore *oc);
-void EXP_Remove(struct objcore *);
+void EXP_Remove(struct objcore *, const struct objcore *);
 
 #define EXP_Dttl(req, oc) (oc->ttl - (req->t_req - oc->t_origin))
 
@@ -303,6 +302,7 @@ unsigned HTTP1_Write(const struct worker *w, const struct http *hp, const int*);
 
 /* cache_main.c */
 vxid_t VXID_Get(const struct worker *, uint64_t marker);
+extern pthread_key_t panic_key;
 extern pthread_key_t witness_key;
 
 void THR_SetName(const char *name);
@@ -435,7 +435,7 @@ void SES_Wait(struct sess *, const struct transport *);
 void SES_Ref(struct sess *sp);
 void SES_Rel(struct sess *sp);
 
-const char * HTC_Status(enum htc_status_e);
+void HTC_Status(enum htc_status_e, const char **, const char **);
 void HTC_RxInit(struct http_conn *htc, struct ws *ws);
 void HTC_RxPipeline(struct http_conn *htc, char *);
 enum htc_status_e HTC_RxStuff(struct http_conn *, htc_complete_f *,
@@ -462,7 +462,8 @@ struct conn_pool;
 void VCP_Init(void);
 void VCP_Panic(struct vsb *, struct conn_pool *);
 
-/* cache_backend_poll.c */
+/* cache_backend_probe.c */
+void VBP_Init(void);
 
 /* cache_vary.c */
 int VRY_Create(struct busyobj *bo, struct vsb **psb);
@@ -563,6 +564,7 @@ const struct stevedore *STV_next(void);
 int STV_BanInfoDrop(const uint8_t *ban, unsigned len);
 int STV_BanInfoNew(const uint8_t *ban, unsigned len);
 void STV_BanExport(const uint8_t *banlist, unsigned len);
+// STV_NewObject() len is space for OBJ_VARATTR
 int STV_NewObject(struct worker *, struct objcore *,
     const struct stevedore *, unsigned len);
 
