@@ -97,7 +97,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	acceptor_sleep_incr,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0",
 	/* max */	"1",
 	/* def */	"0",
@@ -112,7 +112,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	acceptor_sleep_max,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0",
 	/* max */	"10",
 	/* def */	"0.05",
@@ -182,7 +182,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	ban_lurker_age,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0",
 	/* max */	NULL,
 	/* def */	"60",
@@ -211,7 +211,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	ban_lurker_sleep,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0",
 	/* max */	NULL,
 	/* def */	"0.010",
@@ -225,7 +225,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	ban_lurker_holdoff,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0",
 	/* max */	NULL,
 	/* def */	"0.010",
@@ -267,7 +267,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	backend_idle_timeout,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"1",
 	/* max */	NULL,
 	/* def */	"60",
@@ -278,7 +278,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	backend_local_error_holddown,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0.000",
 	/* max */	NULL,
 	/* def */	"10.000",
@@ -296,7 +296,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	backend_remote_error_holddown,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0.000",
 	/* max */	NULL,
 	/* def */	"0.250",
@@ -361,7 +361,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	clock_step,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0.000",
 	/* max */	NULL,
 	/* def */	"1.000",
@@ -387,7 +387,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	critbit_cooloff,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"60.000",
 	/* max */	"254.000",
 	/* def */	"180.000",
@@ -400,7 +400,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	default_grace,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0.000",
 	/* max */	NULL,
 	/* def */	"10s",
@@ -417,7 +417,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	default_keep,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0.000",
 	/* max */	NULL,
 	/* def */	"0s",
@@ -435,7 +435,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	default_ttl,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0.000",
 	/* max */	NULL,
 	/* def */	"2m",
@@ -643,9 +643,9 @@ PARAM_SIMPLE(
 PARAM_SIMPLE(
 	/* name */	idle_send_timeout,
 	/* type */	timeout,
-	/* min */	"0.000",
-	/* max */	NULL,
-	/* def */	"60.000",
+	/* min */	"0s",
+	/* max */	"1h",
+	/* def */	"60s",
 	/* units */	"seconds",
 	/* descr */
 	"Send timeout for individual pieces of data on client connections."
@@ -670,7 +670,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	lru_interval,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0.000",
 	/* max */	NULL,
 	/* def */	"2.000",
@@ -766,6 +766,18 @@ PARAM_SIMPLE(
 	"Maximum number of sessions dedicated to pipe transactions."
 )
 
+PARAM_SIMPLE(
+	/* name */	pipe_task_deadline,
+	/* type */	timeout,
+	/* min */	"0.000",
+	/* max */	NULL,
+	/* def */	"0s",
+	/* units */	"seconds",
+	/* descr */
+	"Deadline for PIPE sessions. Regardless of activity in either "
+	"direction after this many seconds, the session is closed."
+)
+
 PARAM_SIMPLE(
 	/* name */	pipe_timeout,
 	/* type */	timeout,
@@ -824,7 +836,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	shortlived,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0.000",
 	/* max */	NULL,
 	/* def */	"10.000",
@@ -883,7 +895,7 @@ PARAM_SIMPLE(
 #endif
 PARAM_SIMPLE(
 	/* name */	tcp_keepalive_intvl,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"1",
 	/* max */	"100",
 	/* def */	NULL,
@@ -916,7 +928,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	tcp_keepalive_time,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"1",
 	/* max */	"7200",
 	/* def */	NULL,
@@ -934,10 +946,10 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	timeout_idle,
-	/* type */	timeout,
-	/* min */	"0.000",
-	/* max */	NULL,
-	/* def */	"5.000",
+	/* type */	duration,
+	/* min */	"0s",
+	/* max */	"1h",
+	/* def */	"5s",
 	/* units */	"seconds",
 	/* descr */
 	"Idle timeout for client connections.\n\n"
@@ -950,7 +962,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	timeout_linger,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0.000",
 	/* max */	NULL,
 	/* def */	"0.050",
@@ -1005,7 +1017,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	vcl_cooldown,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"1.000",
 	/* max */	NULL,
 	/* def */	"600.000",
@@ -1043,7 +1055,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	vsm_free_cooldown,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"10.000",
 	/* max */	"600.000",
 	/* def */	"60.000",
@@ -1193,6 +1205,20 @@ PARAM_SIMPLE(
 	/* flags */	WIZARD
 )
 
+PARAM_SIMPLE(
+	/* name */	h2_window_timeout,
+	/* type */	timeout,
+	/* min */	"0",
+	/* max */	NULL,
+	/* def */	"5",
+	/* units */	"seconds",
+	/* descr */
+	"HTTP2 time limit without window credits. How long a stream may "
+	"wait for the client to credit the window and allow for more DATA "
+	"frames to be sent.",
+	/* flags */	WIZARD
+)
+
 PARAM_SIMPLE(
 	/* name */	h2_header_table_size,
 	/* type */	bytes_u,
@@ -1263,7 +1289,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	h2_rapid_reset,
-	/* typ */	timeout,
+	/* typ */	duration,
 	/* min */	"0.000",
 	/* max */	NULL,
 	/* def */	"1.000",
@@ -1293,7 +1319,7 @@ PARAM_SIMPLE(
 
 PARAM_SIMPLE(
 	/* name */	h2_rapid_reset_period,
-	/* typ */	timeout,
+	/* typ */	duration,
 	/* min */	"1.000",
 	/* max */	NULL,
 	/* def */	"60.000",
@@ -1446,7 +1472,7 @@ PARAM_THREAD(
 PARAM_THREAD(
 	/* name */	thread_pool_timeout,
 	/* field */	timeout,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"10",
 	/* max */	NULL,
 	/* def */	"300",
@@ -1462,7 +1488,7 @@ PARAM_THREAD(
 PARAM_THREAD(
 	/* name */	thread_pool_watchdog,
 	/* field */	watchdog,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0.1",
 	/* max */	NULL,
 	/* def */	"60",
@@ -1478,7 +1504,7 @@ PARAM_THREAD(
 PARAM_THREAD(
 	/* name */	thread_pool_destroy_delay,
 	/* field */	destroy_delay,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0.01",
 	/* max */	NULL,
 	/* def */	"1",
@@ -1493,7 +1519,7 @@ PARAM_THREAD(
 PARAM_THREAD(
 	/* name */	thread_pool_add_delay,
 	/* field */	add_delay,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	"0",
 	/* max */	NULL,
 	/* def */	"0",
@@ -1513,7 +1539,7 @@ PARAM_THREAD(
 PARAM_THREAD(
 	/* name */	thread_pool_fail_delay,
 	/* field */	fail_delay,
-	/* type */	timeout,
+	/* type */	duration,
 	/* min */	".01",
 	/* max */	NULL,
 	/* def */	"0.2",
@@ -1829,11 +1855,12 @@ PARAM_BITS(
 	/* name */	feature,
 	/* fld */	feature_bits,
 	/* def */
+	"none,"
 	"+validate_headers,"
 	"+vcl_req_reset",
 	/* descr */
 	"Enable/Disable various minor features.\n"
-	"\tdefault\tSet default value\n"
+	"\tdefault\tSet default value (deprecated: use param.reset)\n"
 	"\tnone\tDisable all features.\n\n"
 	"Use +/- prefix to enable/disable individual feature:")
 #ifdef PARAM_ALL
@@ -1847,11 +1874,12 @@ PARAM_BITS(
 	/* name */	vcc_feature,
 	/* fld */	vcc_feature_bits,
 	/* def */
+	"none,"
 	"+err_unref,"
 	"+unsafe_path",
 	/* descr */
 	"Enable/Disable various VCC behaviors.\n"
-	"\tdefault\tSet default value\n"
+	"\tdefault\tSet default value (deprecated: use param.reset)\n"
 	"\tnone\tDisable all behaviors.\n\n"
 	"Use +/- prefix to enable/disable individual behavior:")
 #ifdef PARAM_ALL
@@ -1865,6 +1893,7 @@ PARAM_BITS(
 	/* name */	vsl_mask,
 	/* fld */	vsl_mask,
 	/* def */
+	"all,"
 	"-Debug,"
 	"-ExpKill,"
 	"-H2RxBody,"
@@ -1881,9 +1910,10 @@ PARAM_BITS(
 	"-WorkThread",
 	/* descr */
 	"Mask individual VSL messages from being logged.\n"
-	"\tdefault\tSet default value\n"
+	"\tall\tEnable all tags\n"
+	"\tdefault\tSet default value (deprecated: use param.reset)\n"
 	"\nUse +/- prefix in front of VSL tag name to unmask/mask "
-	"individual VSL messages.")
+	"individual VSL messages. See vsl(7) for possible values.")
 PARAM_POST
 
 #undef PARAM_ALL
