@@ -34,7 +34,7 @@
  * REMEMBER to update the documentation (especially the varnishlog(1) man
  * page) whenever this list changes.
  *
- * XXX: Please add new entries a the end to not break saved log-segments.
+ * XXX: Please add new entries at the end to not break saved log-segments.
  *
  * Arguments:
  *	Tag-Name
@@ -91,6 +91,17 @@ SLTM(SessOpen, 0, "Client connection opened",
 	"\t+------------------- Remote IPv4/6 address / 0.0.0.0 for UDS\n"
 	"\n"
 )
+/*
+ * XXX generate the list of SC_* reasons:
+ *
+ * #include <stdio.h>
+ * int main(void) {
+ * #define SESS_CLOSE(e, sc, err, desc)			\
+ *	printf("\t\"\\t* ``%s``: %s\\n\"\n", #e, desc);
+ * #include "include/tbl/sess_close.h"
+ *	return (0);
+ * }
+ */
 
 SLTM(SessClose, 0, "Client connection closed",
 	"SessClose is the last record for any client connection.\n\n"
@@ -99,6 +110,26 @@ SLTM(SessClose, 0, "Client connection closed",
 	"\t|  |\n"
 	"\t|  +- How long the session was open\n"
 	"\t+---- Why the connection closed\n"
+	"\nExplanation of reasons (first column):\n"
+	"\t* ``REM_CLOSE``: Client Closed\n"
+	"\t* ``REQ_CLOSE``: Client requested close\n"
+	"\t* ``REQ_HTTP10``: Proto < HTTP/1.1\n"
+	"\t* ``RX_BAD``: Received bad req/resp\n"
+	"\t* ``RX_BODY``: Failure receiving body\n"
+	"\t* ``RX_JUNK``: Received junk data\n"
+	"\t* ``RX_OVERFLOW``: Received buffer overflow\n"
+	"\t* ``RX_TIMEOUT``: Receive timeout\n"
+	"\t* ``RX_CLOSE_IDLE``: timeout_idle reached\n"
+	"\t* ``TX_PIPE``: Piped transaction\n"
+	"\t* ``TX_ERROR``: Error transaction\n"
+	"\t* ``TX_EOF``: EOF transmission\n"
+	"\t* ``RESP_CLOSE``: Backend/VCL requested close\n"
+	"\t* ``OVERLOAD``: Out of some resource\n"
+	"\t* ``PIPE_OVERFLOW``: Session pipe overflow\n"
+	"\t* ``RANGE_SHORT``: Insufficient data for range\n"
+	"\t* ``REQ_HTTP20``: HTTP2 not accepted\n"
+	"\t* ``VCL_FAILURE``: VCL failure\n"
+	"\t* ``RAPID_RESET``: HTTP2 rapid reset\n"
 	"\n"
 )
 
@@ -124,7 +155,7 @@ SLTM(BackendClose, 0, "Backend connection closed",
 	"The format is::\n\n"
 	"\t%d %s %s [ %s ]\n"
 	"\t|  |  |    |\n"
-	"\t|  |  |    +- Optional reason\n"
+	"\t|  |  |    +- Optional reason, see SessClose for explanation\n"
 	"\t|  |  +------ \"close\" or \"recycle\"\n"
 	"\t|  +--------- Backend display name\n"
 	"\t+------------ Connection file descriptor\n"
@@ -158,33 +189,9 @@ SLTM(Length, 0, "Size of object body",
 	"Logs the size of a fetch object body.\n\n"
 )
 
-/*
- * XXX generate HTC info below from tbl include
- *
- * #include <stdio.h>
- * int main(void) {
- * #define HTC_STATUS(e, n, s, l) \
- *	printf("\t\"\\t* %s (%d): %s\\n\"\n", s, n, l);
- * #include "include/tbl/htc.h"
- *	return (0);
- * }
- */
-
 SLTM(FetchError, 0, "Error while fetching object",
 	"Logs the error message of a failed fetch operation.\n\n"
-	"Error messages should be self-explanatory, yet the http connection\n"
-	"(HTC) class of errors is reported with these symbols:\n\n"
-	"\t* junk (-5): Received unexpected data\n"
-	"\t* close (-4): Connection closed\n"
-	"\t* timeout (-3): Timed out\n"
-	"\t* overflow (-2): Buffer/workspace too small\n"
-	"\t* eof (-1): Unexpected end of input\n"
-	"\t* empty (0): Empty response\n"
-	"\t* more (1): More data required\n"
-	"\t* complete (2): Data complete (no error)\n"
-	"\t* idle (3): Connection was closed while idle\n"
-	"\nNotice that some HTC errors are never emitted."
-	)
+)
 
 #define SLTH(tag, ind, req, resp, sdesc, ldesc) \
 	SLTM(Req##tag, (req ? 0 : SLT_F_UNUSED), "Client request " sdesc, ldesc)
@@ -264,15 +271,19 @@ SLTM(Fetch_Body, 0, "Body fetched from backend",
 SLTM(VCL_acl, 0, "VCL ACL check results",
 	"ACLs with the `+log` flag emits this record with the result.\n\n"
 	"The format is::\n\n"
-	"\t%s %s [%s [fixed: %s]]\n"
-	"\t|  |   |          |\n"
-	"\t|  |   |          +- Fixed entry (see acl +pedantic flag)\n"
-	"\t|  |   +------------ Matching entry (only for MATCH)\n"
-	"\t|  +---------------- Name of the ACL\n"
-	"\t+-------------------- MATCH or NO_MATCH\n"
+	"\t%s [%s [%s [fixed: %s]]]\n"
+	"\t|   |   |          |\n"
+	"\t|   |   |          +- Fix info (see below)\n"
+	"\t|   |   +------------ Matching entry (only for MATCH)\n"
+	"\t|   +---------------- Name of the ACL for MATCH or NO_MATCH\n"
+	"\t+-------------------- MATCH, NO_MATCH or NO_FAM\n"
 	"\n"
-	"MATCH denotes an ACL match\n"
-	"NO_MATCH denotes that a checked ACL has not matched\n"
+	"* Fix info: either contains network/mask for non-canonical entries "
+	"(see acl +pedantic flag) or ``folded`` for entries "
+	"which were the result of a fold operation (see acl +fold flag).\n"
+	"* ``MATCH`` denotes an ACL match\n"
+	"* ``NO_MATCH`` denotes that a checked ACL has not matched\n"
+	"* ``NO_FAM`` denotes a missing address family and should not occur.\n"
 	"\n"
 )
 
@@ -349,8 +360,15 @@ SLTM(ExpKill, 0, "Object expiry event",
 	"\tLogged when the expiry thread kills an object from the inbox.\n\n"
 	"EXP_When\n"
 	"\tLogged when the expiry thread moves an object on the binheap.\n\n"
+	"EXP_Inspect\n"
+	"\tLogged when the expiry thread inspects the next object scheduled"
+	" to expire.\n\n"
 	"EXP_Expired\n"
 	"\tLogged when the expiry thread expires an object.\n\n"
+	"EXP_Removed\n"
+	"\tLogged when the expiry thread removes an object before expiry.\n\n"
+	"VBF_Superseded\n"
+	"\tLogged when an object supersedes another.\n\n"
 	"LRU_Cand\n"
 	"\tLogged when an object is evaluated for LRU force expiry.\n\n"
 	"LRU\n"
@@ -363,7 +381,10 @@ SLTM(ExpKill, 0, "Object expiry event",
 	"\tEXP_Inbox p=%p e=%f f=0x%x\n"
 	"\tEXP_Kill p=%p e=%f f=0x%x\n"
 	"\tEXP_When p=%p e=%f f=0x%x\n"
-	"\tEXP_Expired x=%u t=%f\n"
+	"\tEXP_Inspect p=%p e=%f f=0x%x\n"
+	"\tEXP_Expired x=%u t=%f h=%u\n"
+	"\tEXP_Removed x=%u t=%f h=%u\n"
+	"\tVBF_Superseded x=%u n=%u\n"
 	"\tLRU_Cand p=%p f=0x%x r=%d\n"
 	"\tLRU x=%u\n"
 	"\tLRU_Fail\n"
@@ -376,6 +397,8 @@ SLTM(ExpKill, 0, "Object expiry event",
 	"\tf=0x%x       Objcore flags\n"
 	"\tr=%d         Objcore refcount\n"
 	"\tx=%u         Object VXID\n"
+	"\tn=%u         New object VXID\n"
+	"\th=%u         Objcore hits\n"
 	"\n"
 )
 
